use crate::playlist::Playlist;
use crate::track::Track;
use reqwest::Client;
use serde::{Deserialize, Serialize};
use std::fmt::{Display, Formatter};
use thiserror::Error;
use uuid::Uuid;

#[derive(Clone)]
pub struct Subsonic {
    username: String,
    password: String,
    base: String,
    client: Client,
}

#[derive(Serialize, Deserialize)]
pub struct SubsonicResponse<T> {
    #[serde(rename = "subsonic-response")]
    subsonic_response: T,
}

#[derive(Serialize, Deserialize)]
struct SearchResponse {
    #[serde(rename = "searchResult3")]
    search_results: SearchResponseSongs,
}

#[derive(Serialize, Deserialize)]
struct SearchResponseSongs {
    song: Vec<Track>,
}

#[derive(Serialize, Deserialize)]
struct GetPlaylistResponse {
    playlist: Playlist,
}

type CreatePlaylistResponse = GetPlaylistResponse;

struct AuthParams {
    salt: String,
    hash: String,
}

#[derive(Debug, Error)]
pub enum SubsonicError {
    Reqwest(reqwest::Error),
}

impl Display for SubsonicError {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "{:?}", self)
    }
}

impl Subsonic {
    fn auth_params(&self) -> AuthParams {
        let salt = Uuid::new_v4().simple().to_string();
        let hash = format!(
            "{:?}",
            md5::compute(format!("{}{}", self.password, salt).into_bytes())
        );

        AuthParams { salt, hash }
    }

    pub async fn song_search(&self, query: String) -> Result<Vec<Track>, SubsonicError> {
        let auth_params = self.auth_params();
        let query_params = [
            ("query", query.as_str()),
            ("artistCount", "0"),
            ("albumCount", "0"),
            ("u", self.username.as_str()),
            ("s", auth_params.salt.as_str()),
            ("t", auth_params.hash.as_str()),
            ("v", "16"),
            ("c", "playlistd"),
            ("f", "json"),
        ];

        Ok(self
            .client
            .get(format!("{}/search3", self.base))
            .query(&query_params)
            .send()
            .await
            .map_err(|e| SubsonicError::Reqwest(e))?
            .json::<SubsonicResponse<SearchResponse>>()
            .await
            .map_err(|e| SubsonicError::Reqwest(e))?
            .subsonic_response
            .search_results
            .song)
    }

    pub async fn get_playlist(&self, playlist_id: String) -> Result<Playlist, SubsonicError> {
        let auth_params = self.auth_params();

        let get_params = [
            ("u", self.username.to_string()),
            ("s", auth_params.salt.to_string()),
            ("t", auth_params.hash.to_string()),
            ("v", String::from("16")),
            ("c", String::from("playlistd")),
            ("f", String::from("json")),
            ("id", playlist_id),
        ];

        Ok(self
            .client
            .get(format!("{}/getPlaylist", self.base))
            .query(&get_params)
            .send()
            .await
            .map_err(|e| SubsonicError::Reqwest(e))?
            .json::<SubsonicResponse<GetPlaylistResponse>>()
            .await
            .map_err(|e| SubsonicError::Reqwest(e))?
            .subsonic_response
            .playlist)
    }

    pub async fn create_playlist(&self, playlist_name: String) -> Result<Playlist, SubsonicError> {
        let auth_params = self.auth_params();

        let get_params = [
            ("u", self.username.to_string()),
            ("s", auth_params.salt.to_string()),
            ("t", auth_params.hash.to_string()),
            ("v", String::from("16")),
            ("c", String::from("playlistd")),
            ("f", String::from("json")),
            ("name", playlist_name),
        ];

        Ok(self
            .client
            .get(format!("{}/createPlaylist", self.base))
            .query(&get_params)
            .send()
            .await
            .map_err(|e| SubsonicError::Reqwest(e))?
            .json::<SubsonicResponse<CreatePlaylistResponse>>()
            .await
            .map_err(|e| SubsonicError::Reqwest(e))?
            .subsonic_response
            .playlist)
    }

    pub async fn update_playlist(
        &self,
        playlist_id: String,
        add: Vec<String>,
        remove: Vec<usize>,
    ) -> Result<Playlist, SubsonicError> {
        let auth_params = self.auth_params();

        let mut update_query = vec![
            ("u", self.username.to_string()),
            ("s", auth_params.salt.to_string()),
            ("t", auth_params.hash.to_string()),
            ("v", String::from("16")),
            ("c", String::from("playlistd")),
            ("f", String::from("json")),
            ("playlistId", playlist_id.clone()),
        ];

        for id in add {
            update_query.push(("songIdToAdd", id));
        }

        for index in remove {
            update_query.push(("songIndexToRemove", index.to_string()));
        }

        self.client
            .get(format!("{}/updatePlaylist", self.base))
            .query(&update_query)
            .send()
            .await
            .map_err(|e| SubsonicError::Reqwest(e))?;

        self.get_playlist(playlist_id).await
    }
}

#[derive(Default)]
pub struct SubsonicBuilder {
    username: Option<String>,
    password: Option<String>,
    base: Option<String>,
}

#[derive(Debug, Error)]
enum SubsonicBuilderError {
    #[error("Field not provided")]
    MissingField,
}

impl SubsonicBuilder {
    pub fn new() -> Self {
        Self {
            ..Default::default()
        }
    }

    pub fn username(mut self, username: String) -> Self {
        self.username = Some(username);
        self
    }

    pub fn password(mut self, password: String) -> Self {
        self.password = Some(password);
        self
    }

    pub fn base(mut self, base: String) -> Self {
        self.base = Some(base);
        self
    }

    pub fn build(self) -> Result<Subsonic, Box<dyn std::error::Error>> {
        Ok(Subsonic {
            username: self.username.ok_or(SubsonicBuilderError::MissingField)?,
            password: self.password.ok_or(SubsonicBuilderError::MissingField)?,
            base: self.base.ok_or(SubsonicBuilderError::MissingField)?,
            client: Default::default(),
        })
    }
}
